from myorg import pack_a
from myorg.pack_b import hello_world


def test_package():
    assert hello_world() == "Hello world from package B!"


def test_integration():
    assert pack_a.hello_world() == "Hello world from package A!"



from __future__ import annotations

from pathlib import Path

from tach.constants import CONFIG_FILE_NAME


def build_project_config_path(root: Path) -> Path:
    return root / f"{CONFIG_FILE_NAME}.toml"


def get_project_config_path(root: Path) -> Path | None:
    file_path = build_project_config_path(root)
    if file_path.exists():
        return file_path
    return None


def get_deprecated_project_config_path(root: Path | None = None) -> Path | None:
    root = root or Path.cwd()
    file_path = root / f"{CONFIG_FILE_NAME}.yaml"
    if file_path.exists():
        return file_path
    file_path = root / f"{CONFIG_FILE_NAME}.yml"
    if file_path.exists():
        return file_path
    return None


def find_project_config_root() -> Path | None:
    cwd = Path.cwd()

    if get_project_config_path(cwd) is not None:
        return cwd

    # Iterate upwards, looking for project config
    for parent in cwd.parents:
        if get_project_config_path(parent):
            return parent

    return None



from pathlib import Path

from setuptools import find_namespace_packages, setup


# For this trick, I thank:
# https://stackoverflow.com/a/78056725

root: Path = Path(__file__).parent
pack_b_path: str = (root.parent / "pack-b").as_uri()

if __name__ == "__main__":
    setup(
        install_requires=[f"myorg-pack-b @ {pack_b_path}"],
        packages=find_namespace_packages(include=["myorg.*"]),
    )



from __future__ import annotations

from pathlib import Path
from unittest.mock import Mock

import pytest

from tach import cli
from tach.extension import ProjectConfig


@pytest.fixture
def mock_check(mocker) -> Mock:
    mock = Mock(return_value=[])  # default to a return with no errors
    mocker.patch("tach.cli.check", mock)
    return mock


@pytest.fixture
def mock_project_config(mocker) -> ProjectConfig:
    def _mock_project_config(root: str = "") -> ProjectConfig:
        return ProjectConfig()

    mocker.patch("tach.cli.parse_project_config", _mock_project_config)
    return _mock_project_config()


def test_execute_with_config(capfd, mock_check, mock_project_config):
    # Test with a valid path as mocked
    args, _ = cli.parse_arguments(["check"])
    assert args.command == "check"
    with pytest.raises(SystemExit) as sys_exit:
        cli.tach_check(
            project_root=Path(),
            project_config=mock_project_config,
        )
    captured = capfd.readouterr()
    assert sys_exit.value.code == 0
    assert "✅" in captured.out
    assert "All modules validated!" in captured.out


def test_invalid_command(capfd):
    with pytest.raises(SystemExit) as sys_exit:
        # Test with an invalid command
        cli.parse_arguments(["help"])
    captured = capfd.readouterr()
    assert sys_exit.value.code == 2
    assert "invalid choice: 'help" in captured.err


def test_execute_with_valid_exclude(capfd, mock_check, mock_project_config):
    with pytest.raises(SystemExit) as sys_exit:
        # Test with a valid path as mocked
        args, _ = cli.parse_arguments(["check", "--exclude", "valid_dir"])
        mock_project_config.exclude = args.exclude.split(",")
        cli.tach_check(
            project_root=Path(),
            project_config=mock_project_config,
        )
    captured = capfd.readouterr()
    assert sys_exit.value.code == 0
    assert "✅" in captured.out
    assert "All modules validated!" in captured.out



from __future__ import annotations

from unittest.mock import patch

import pytest

from tach.extension import ProjectConfig
from tach.interactive import InteractiveModuleConfiguration
from tach.mod import mod_edit_interactive
from tach.parsing import parse_project_config


@pytest.fixture
def temp_project_dir(tmp_path):
    """Create a temporary project directory with initial structure."""
    project_root = tmp_path / "test_project"
    project_root.mkdir()
    return project_root


@pytest.fixture
def initial_project_config():
    """Create a clean project config."""
    return ProjectConfig()


def test_mod_edit_interactive_new_configuration(
    temp_project_dir, initial_project_config
):
    mock_config = InteractiveModuleConfiguration(
        source_roots=[temp_project_dir / "src"],
        module_paths=[
            temp_project_dir / "src" / "module1.py",
            temp_project_dir / "src" / "module2.py",
        ],
        utility_paths=[temp_project_dir / "src" / "utils.py"],
    )

    with patch(
        "tach.mod.get_selected_modules_interactive", return_value=mock_config
    ) as mock_get:
        success, errors = mod_edit_interactive(
            project_root=temp_project_dir,
            project_config=initial_project_config,
            exclude_paths=[],
        )

        assert success is True
        assert not errors
        mock_get.assert_called_once()

        # Verify the saved configuration
        saved_config = parse_project_config(temp_project_dir)
        assert set(saved_config.source_roots) == {"src"}
        assert set(saved_config.module_paths()) == {"module1", "module2", "utils"}
        assert set(saved_config.utility_paths()) == {"utils"}


def test_mod_edit_interactive_validation_error(
    temp_project_dir, initial_project_config
):
    # Create a configuration where a module is outside the source roots
    mock_config = InteractiveModuleConfiguration(
        source_roots=[temp_project_dir / "src"],
        module_paths=[temp_project_dir / "outside" / "module1.py"],  # Outside src
        utility_paths=[],
    )

    with patch(
        "tach.mod.get_selected_modules_interactive", return_value=mock_config
    ) as mock_get:
        success, errors = mod_edit_interactive(
            project_root=temp_project_dir,
            project_config=initial_project_config,
            exclude_paths=[],
        )

        assert success is False
        assert len(errors) == 1
        assert "not contained within any source root" in errors[0]
        mock_get.assert_called_once()


def test_mod_edit_interactive_user_cancelled(temp_project_dir, initial_project_config):
    with patch(
        "tach.mod.get_selected_modules_interactive", return_value=None
    ) as mock_get:
        success, errors = mod_edit_interactive(
            project_root=temp_project_dir,
            project_config=initial_project_config,
            exclude_paths=[],
        )

        assert success is False
        assert len(errors) == 1
        assert "No changes saved" in errors[0]
        mock_get.assert_called_once()


def test_mod_edit_interactive_update_existing(temp_project_dir, initial_project_config):
    # First round of edits
    mock_config = InteractiveModuleConfiguration(
        source_roots=[temp_project_dir / "src", temp_project_dir / "tests"],
        module_paths=[
            temp_project_dir / "src" / "new_module.py",
        ],
        utility_paths=[
            temp_project_dir / "src" / "new_utility.py",
        ],
    )

    with patch(
        "tach.mod.get_selected_modules_interactive", return_value=mock_config
    ) as mock_get:
        success, errors = mod_edit_interactive(
            project_root=temp_project_dir,
            project_config=initial_project_config,
            exclude_paths=[],
        )

        assert success is True
        assert not errors
        mock_get.assert_called_once()

        saved_config = parse_project_config(temp_project_dir)
        assert set(saved_config.source_roots) == {"src", "tests"}
        assert set(saved_config.module_paths()) == {"new_module", "new_utility"}
        assert set(saved_config.utility_paths()) == {"new_utility"}

    # Second round of edits
    mock_config_2 = InteractiveModuleConfiguration(
        source_roots=[temp_project_dir / "src"],  # Remove tests directory
        module_paths=[
            temp_project_dir / "src" / "new_module.py",
            temp_project_dir / "src" / "another_module.py",  # Add new module
        ],
        utility_paths=[],  # Remove utility
    )

    with patch(
        "tach.mod.get_selected_modules_interactive", return_value=mock_config_2
    ) as mock_get:
        success, errors = mod_edit_interactive(
            project_root=temp_project_dir,
            project_config=saved_config,  # Use previously saved config
            exclude_paths=[],
        )

        assert success is True
        assert not errors
        mock_get.assert_called_once()

        # Verify final configuration
        final_config = parse_project_config(temp_project_dir)
        assert set(final_config.source_roots) == {"src"}
        assert set(final_config.module_paths()) == {"new_module", "another_module"}
        assert set(final_config.utility_paths()) == set()



from __future__ import annotations

from tach.interactive.modules import (
    InteractiveModuleConfiguration,
    get_selected_modules_interactive,
)

__all__ = ["get_selected_modules_interactive", "InteractiveModuleConfiguration"]



from __future__ import annotations

from tach.parsing.config import (
    dump_project_config_to_toml,
    extend_and_validate,
    parse_project_config,
)

__all__ = ["parse_project_config", "dump_project_config_to_toml", "extend_and_validate"]



from __future__ import annotations

import pytest

from tach.parsing import parse_project_config


def test_invalid_project_config(example_dir):
    with pytest.raises(ValueError):
        parse_project_config(example_dir / "invalid")


def test_empty_project_config(example_dir):
    with pytest.raises(ValueError):
        parse_project_config(example_dir / "invalid" / "empty")



from pathlib import Path

from setuptools import setup


# For this trick, I thank:
# https://stackoverflow.com/a/78056725

root: Path = Path(__file__).parent
pack_a_path: str = (root.parent / "pack-a").as_uri()

if __name__ == "__main__":
    setup(
        install_requires=[f"myorg-pack-a @ {pack_a_path}"],
    )



from __future__ import annotations

import json
from json.decoder import JSONDecodeError
from typing import TYPE_CHECKING
from urllib import error, request

from tach.constants import GAUGE_API_BASE_URL
from tach.extension import serialize_modules_json

if TYPE_CHECKING:
    from pathlib import Path

    import pydot  # type: ignore

    from tach.extension import ProjectConfig


def generate_show_url(
    project_config: ProjectConfig,
    included_paths: list[Path],
) -> str | None:
    modules = project_config.filtered_modules(included_paths)
    for module in modules:
        if module.depends_on is None:
            # This is a hack to avoid bumping the API version
            module.depends_on = []
    json_data = serialize_modules_json(modules)
    json_bytes = json_data.encode("utf-8")
    req = request.Request(
        f"{GAUGE_API_BASE_URL}/api/show/graph/1.3",
        data=json_bytes,
        headers={"Content-Type": "application/json"},
    )

    try:
        # Send the request and read the response
        with request.urlopen(req) as response:
            response_data = response.read().decode("utf-8")
            response_json = json.loads(response_data)
            uid = response_json.get("uid")
            return f"{GAUGE_API_BASE_URL}/show?uid={uid}"
    except (UnicodeDecodeError, JSONDecodeError, error.URLError) as e:
        print(f"Error: {e}")
        return None


def generate_module_graph_dot_file(
    project_config: ProjectConfig,
    output_filepath: Path,
    included_paths: list[Path],
) -> None:
    # Local import because networkx takes about ~100ms to load
    import networkx as nx

    graph = nx.DiGraph()  # type: ignore

    def upsert_edge(graph: nx.DiGraph, module: str, dependency: str) -> None:  # type: ignore
        if module not in graph:
            graph.add_node(module)  # type: ignore
        if dependency not in graph:
            graph.add_node(dependency)  # type: ignore
        graph.add_edge(module, dependency)  # type: ignore

    modules = project_config.filtered_modules(included_paths)

    for module in modules:
        for dependency in module.depends_on or []:
            upsert_edge(graph, module.path, dependency.path)  # type: ignore

    pydot_graph: pydot.Dot = nx.nx_pydot.to_pydot(graph)  # type: ignore
    dot_data: str = pydot_graph.to_string()  # type: ignore

    output_filepath.write_text(dot_data)  # type: ignore


def generate_module_graph_mermaid(
    project_config: ProjectConfig,
    output_filepath: Path,
    included_paths: list[Path],
) -> None:
    modules = project_config.filtered_modules(included_paths)
    edges: list[str] = []
    isolated: list[str] = []
    for module in modules:
        for dependency in module.depends_on or []:
            edges.append(
                f"    {module.path.strip('<>')} --> {dependency.path.strip('<>')}"
            )
        if not module.depends_on:
            isolated.append(f"    {module.path.strip('<>')}")

    mermaid_graph = "graph TD\n" + "\n".join(edges) + "\n" + "\n".join(isolated)

    output_filepath.write_text(mermaid_graph)


__all__ = [
    "generate_show_url",
    "generate_module_graph_dot_file",
    "generate_module_graph_mermaid",
]



from __future__ import annotations

import json
import os
import re
from dataclasses import asdict, dataclass, field
from http.client import HTTPConnection, HTTPSConnection
from typing import TYPE_CHECKING, Any
from urllib import parse

from tach import filesystem as fs
from tach.colors import BCOLORS
from tach.constants import GAUGE_API_BASE_URL
from tach.errors import TachClosedBetaError, TachError
from tach.extension import (
    ProjectConfig,
    check,
    get_project_imports,
    into_usage_errors,
)
from tach.extension import (
    UsageError as ExtUsageError,
)
from tach.filesystem.git_ops import get_current_branch_info
from tach.parsing import extend_and_validate

if TYPE_CHECKING:
    try:
        # dataclass reads the annotation but doesn't need to evaluate it (just checking for ClassVar etc.)
        from typing import Literal  # noqa: TC004
    except ImportError:
        pass
    from pathlib import Path


def export_report(
    project_root: Path,
    project_config: ProjectConfig,
    output_path: Path | None = None,
    force: bool = False,
):
    """Export a modularity report to a local file."""
    report = generate_modularity_report(project_root, project_config, force=force)
    output_path = output_path or project_root / "modularity_report.json"
    output_path.write_text(json.dumps(asdict(report), indent=2))


def upload_report_to_gauge(
    project_root: Path,
    project_config: ProjectConfig,
    force: bool = False,
):
    """Upload a modularity report to Gauge."""
    report = generate_modularity_report(project_root, project_config, force=force)
    print(f"{BCOLORS.OKCYAN} > Uploading report...{BCOLORS.ENDC}")
    response_data = post_json_to_gauge_api(asdict(report))
    print(f"{BCOLORS.OKGREEN} > Report uploaded!{BCOLORS.ENDC}")
    if response_data.get("url"):
        print(
            f"{BCOLORS.OKBLUE} > {GAUGE_API_BASE_URL}{response_data['url']}{BCOLORS.ENDC}"
        )


GAUGE_API_KEY = os.getenv("GAUGE_API_KEY", "")
GAUGE_UPLOAD_URL = f"{GAUGE_API_BASE_URL}/api/client/tach-upload/1.4"


def post_json_to_gauge_api(
    data: dict[str, Any], user_name: str | None = None
) -> dict[str, str]:
    if not GAUGE_API_KEY:
        raise TachClosedBetaError(
            f"{BCOLORS.WARNING}Modularity is currently in closed beta. Visit {GAUGE_API_BASE_URL}/closed-beta to request access.{BCOLORS.ENDC}"
            "\n\n"
            f"{BCOLORS.OKCYAN}Already have access? Set the GAUGE_API_KEY environment variable to continue.{BCOLORS.ENDC}"
        )
    headers = {
        "Content-Type": "application/json",
        "Authorization": GAUGE_API_KEY,
    }
    if user_name:
        data["user_name"] = user_name
    json_data = json.dumps(data)
    conn = None
    try:
        url_parts: parse.ParseResult = parse.urlparse(GAUGE_UPLOAD_URL)
        if GAUGE_UPLOAD_URL.startswith("https://"):
            conn = HTTPSConnection(url_parts.netloc, timeout=10)
        else:
            conn = HTTPConnection(url_parts.netloc, timeout=10)
        conn.request("POST", url_parts.path, body=json_data, headers=headers)
        response = conn.getresponse()
        response_data = response.read().decode("utf-8")
        # If key is unbound, prompt user to provide username to bind key
        if response.status == 422 and not user_name:
            # Prompt user to provide username
            conn.close()
            user_name = input("Enter your GitHub username: ").strip()
            return post_json_to_gauge_api(data, user_name)
        if response.status != 200:
            raise TachError(
                f"API request failed with status {response.status}: {response_data}"
            )
    except Exception as e:
        raise TachError(f"Failed to upload modularity report: {str(e)}")
    finally:
        if conn is not None:
            conn.close()
    return json.loads(response_data)


# NOTE: these usages are all imports
@dataclass
class Usage:
    # The configured Module being used
    module_path: str
    # The full import path to the member being used
    full_path: str
    # The file containing the usage
    filepath: str
    # 1-indexed location of the usage in the file
    line: int
    # [1.1] The Module that contains the Usage (None indicates Usage outside of any Module)
    containing_module_path: str | None


@dataclass
class Dependency:
    path: str
    deprecated: bool = False


@dataclass
class Module:
    path: str
    # [1.2] Replaces 'is_strict'
    has_interface: bool = False
    interface_members: list[str] = field(default_factory=list)
    # [1.3] Adds 'depends_on'
    depends_on: list[Dependency] = field(default_factory=list)


REPORT_VERSION = "1.4"


@dataclass
class ReportMetadata:
    version: str = REPORT_VERSION
    configuration_format: str = "json"


# This is unfortunately necessary because we use 'asdict' on the report
# This will be removed once we move report generation to Rust
@dataclass
class UsageError:
    file: str
    line_number: int
    member: str
    usage_module: str
    definition_module: str
    error_type: Literal["DEPENDENCY", "INTERFACE"]

    @classmethod
    def from_extension(cls, ext_usage_error: ExtUsageError) -> UsageError:
        return cls(
            file=ext_usage_error.file,
            line_number=ext_usage_error.line_number,
            member=ext_usage_error.member,
            usage_module=ext_usage_error.usage_module,
            definition_module=ext_usage_error.definition_module,
            error_type=ext_usage_error.error_type,
        )


@dataclass
class Report:
    email: str
    user_name: str
    owner: str
    repo: str
    branch: str
    commit: str
    # [1.1] The full configuration encoded as JSON
    full_configuration: str
    modules: list[Module] = field(default_factory=list)
    usages: list[Usage] = field(default_factory=list)
    # [1.4] Diagnostics
    diagnostics: list[UsageError] = field(default_factory=list)
    metadata: ReportMetadata = field(default_factory=ReportMetadata)


def build_modules(project_config: ProjectConfig) -> list[Module]:
    modules: list[Module] = []
    for module in project_config.all_modules():
        if module.mod_path() == ".":
            # Skip <root>
            continue

        has_interface = False
        interface_members: set[str] = set()
        for interface in project_config.all_interfaces():
            if any(
                re.match(r"^" + pattern + r"$", module.path)
                for pattern in interface.from_modules
            ):
                has_interface = True
                interface_members.update(interface.expose)
        dependencies = [
            Dependency(path=dep.path, deprecated=dep.deprecated)
            for dep in module.depends_on
            or []  # TODO: platform should differentiate None vs. []
        ]
        modules.append(
            Module(
                path=module.path,
                has_interface=has_interface,
                interface_members=list(interface_members),
                depends_on=dependencies,
            )
        )
    return modules


def build_usages(
    project_root: Path, source_roots: list[Path], project_config: ProjectConfig
) -> list[Usage]:
    module_paths = sorted(
        (module.path for module in project_config.all_modules()),
        key=lambda path: len(path.split(".")),
        reverse=True,
    )

    def get_containing_module(mod_path: str) -> str | None:
        return next(
            (
                module_path
                for module_path in module_paths
                if mod_path == module_path or mod_path.startswith(f"{module_path}.")
            ),
            None,
        )

    usages: list[Usage] = []
    for pyfile in fs.walk_pyfiles(
        project_root,
        project_root=project_root,
        exclude_paths=project_config.exclude,
        use_regex_matching=project_config.use_regex_matching,
    ):
        pyfile_mod_path = fs.file_to_module_path(
            tuple(source_roots), project_root / pyfile
        )
        pyfile_containing_module = get_containing_module(pyfile_mod_path)
        imports = get_project_imports(
            project_root=project_root,
            source_roots=source_roots,
            file_path=pyfile,
            project_config=project_config,
        )
        for project_import in imports:
            import_containing_module = get_containing_module(project_import.module_path)
            if (
                import_containing_module is None
                or import_containing_module == pyfile_containing_module
            ):
                continue

            usages.append(
                Usage(
                    module_path=import_containing_module,
                    full_path=project_import.module_path,
                    filepath=str(pyfile),
                    line=project_import.line_number,
                    containing_module_path=pyfile_containing_module,
                )
            )

    return usages


def generate_modularity_report(
    project_root: Path, project_config: ProjectConfig, force: bool = False
) -> Report:
    print(f"{BCOLORS.OKCYAN} > Generating report...{BCOLORS.ENDC}")
    branch_info = get_current_branch_info(project_root, allow_dirty=force)
    report = Report(
        user_name="",  # only needed for binding a new api key
        email=branch_info.email,
        owner=branch_info.owner,
        repo=branch_info.repo,
        branch=branch_info.name,
        commit=branch_info.commit,
        full_configuration=project_config.serialize_json(),
    )
    source_roots = [project_root / root for root in project_config.source_roots]

    report.modules = build_modules(project_config)
    report.usages = build_usages(project_root, source_roots, project_config)
    project_config.exclude = extend_and_validate(
        None, project_config.exclude, project_config.use_regex_matching
    )
    check_diagnostics = check(
        project_root=project_root,
        project_config=project_config,
        dependencies=True,
        interfaces=True,
    )
    report.diagnostics = list(
        map(
            lambda ext_usage_error: UsageError.from_extension(ext_usage_error),
            into_usage_errors(check_diagnostics),
        )
    )
    print(f"{BCOLORS.OKGREEN} > Report generated!{BCOLORS.ENDC}")
    return report


__all__ = ["export_report", "upload_report_to_gauge"]
